/*
 * MAX : adc224.cc, Wed Jan  2 14:44:59 CET 2002 -sg
 *
 * Device driver for the LeCroy 12 channel CAMAC ADC 2249 SG
 * each channel has his own device file
 * they all use the same initfile
 *
 * C.Rosen, Nov/97 upgrade to 2.0beta
 * 15.12.1998 error()-handling changed (only 0 returns)
 * upgrade to 2.1beta
 * 03.06.1999 set DEBUG-define to unique ident
 * 
 * This file is part of Max data acquisition software
 * (c) 1997-98 by C. Otto & M. Dammasch
 *
 * Max is free software; you can redistribute it and/or modify it
 * under the terms of the version 2 of GNU General Public License as
 * published by the Free Software Foundation.
 *
 * Max is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * (see the file COPYING) in this directory; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "list.hh"
#include "sched.hh"
#include "camacs.h"
#include "error.hh"

#undef DEBUG_DEV_ADC2249

// global variables
static int  N, qx, TLAM, slowmode, chanmax;
static int  *shots, *enabled;

/******************* function(s) common to all channels **********************/
int adc2249_enable ( int channel, Scheduler *parameters )
{
  unsigned int  dummy;

  qx = camac(26,channel,N,&dummy);
  qx=  camac( 9,channel,N,&dummy);

#ifdef DEBUG_DEV_ADC2249
  printf("\nadc2249_%d: enabled",channel);
#endif

  return 1;
}

int adc2249_readout ( int channel, Scheduler *parameters )
{
  int           i, d;
  char          text[21];
  unsigned int  dummy;

  i = 0;
  qx = camac(8,channel,N,&dummy);
  while (++i < TLAM && !(qx & 2))
  {
    if (slowmode) usleep(1);
    qx = camac(8,channel,N,&dummy);
  }

#ifdef DEBUG_DEV_ADC2249
  printf("\nadc2249_%d: qx %d, TLAM %d/%d",channel,qx,i,TLAM);	
#endif
	
  if (i >= TLAM)
  {
    sprintf(text,"LeCroy-ADC2249_%d",channel);
    if (!parameters->error(E_TIMEOUT,text))
      return 0;   //return FALSE -> measurement stops
  }

  qx = camac(0,channel,N,&dummy);   //readout adc
  d = (int) dummy;

  parameters->store((parameters->act_channel() % shots[channel] == 0)?(d):(-1));  //save measured or no-meas value

#ifdef DEBUG_DEV_ADC2249
  printf("\nadc2249_%d: %s data stored",channel,(parameters->act_channel() % shots[channel] == 0)?("real"):("virtual"));
#endif

  return 1;
}

int adc2249_disable ( int channel, Scheduler *parameters )
{
  unsigned int  dummy;

  qx = camac(24,channel,N,&dummy);

#ifdef DEBUG_DEV_ADC2249
  printf("\nadc2249_%d: disabled",channel);
#endif

  return 1;
}
/*****************************************************************************/

//init function
//called each time before meausurement starts
int adc2249_init(List *device_data, Scheduler *parameters)
{	
  unsigned int x;
  char label[12];
  int i;

  if (camac_init() != 1)
    if (!parameters->error(E_CAMAC,"Couldn't initialize device "))
      return 0;

  printf("\ninitializing adc2249");
  device_data->getvar(&N,"adc2249","N");
  device_data->getvar(&slowmode,"adc2249","slow_mode");
  device_data->getvar(&chanmax,"adc2249","channelmax");
  device_data->getvar(&TLAM,"all","timeout");

  shots = (int *)calloc(chanmax,sizeof(int));
  enabled = (int *)calloc(chanmax,sizeof(int));

  //set defaults
  memset(shots,1,sizeof(shots));
  memset(enabled,0,sizeof(enabled));

  for (i = 0; i < chanmax; i++)
  {
    sprintf(label,"shots_ch%d",i);
    device_data->getvar(&shots[i],"adc2249",label);
  }

  for (i = 0; i < chanmax; i++)
  {
    sprintf(label,"channel_%d",i);
    enabled[i] = device_data->enabled("adc2249",label);
  }

  for (i = 0; i < chanmax; i++) {  //clear all(!) channels
    camac(26,i,N,&x);         //enable
    camac(9,i,N,&x);          //erase
    camac(24,i,N,&x);         //disable
  }

  return 1;
}

int adc2249_pre ( List *device_data, Scheduler *parameters )
{
  int i;

  for (i = 0; i < chanmax; i++) {
    if (enabled[i])
      adc2249_enable(i,parameters);
  }

  return 1;
}

int adc2249 ( List *device_data, Scheduler *parameters )
{
  int i;

  for (i = 0; i < chanmax; i++) {
    if (enabled[i]) {
      if ( !(adc2249_readout(i,parameters) && adc2249_disable(i,parameters)))
	return 0;
    }
  }
  return 1;
}

int adc2249_close(List *device_data, Scheduler *parameters)
{
  printf("\nclosing adc2249 ");
  free(shots);
  free(enabled);
  if (camac_close() != 1)
    if (!parameters->error(E_CAMAC,"Couldn't close device"))
      return 0;

  return 1; //function has to return a value -> 1 otherwise measurement stops
}

