/*
 * MAX : lsc4434.cc, Wed Jan  2 14:44:59 CET 2002 -sg
 *
 * Device driver for the LeCroy 32-channel 24bit latching scaler 4434
 * LeCroy Model 4434 MOD 100 TTL inputs version #82530
 *
 * This file is part of Max data acquisition software
 * (c) 1999 by M. Dammasch
 *
 * Max is free software; you can redistribute it and/or modify it
 * under the terms of the version 2 of GNU General Public License as
 * published by the Free Software Foundation.
 *
 * Max is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * (see the file COPYING) in this directory; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "list.hh"
#include "sched.hh"
#include "camacs.h"
#include "error.hh"

#undef DEBUG_DEV_LSC4434

// LSC4434 command register flags
#define TEST  		 (1<<15)// T  bit (w16)
#define NO_TEST 	 0
#define AUXBUS_DISABLE   (1<<13)// BD bit (w14)
#define AUXBUS_ENABLE    0
#define READOUT_ENABLE   (1<<7)// RD bit (w08)
#define READOUT_DISABLE  0
#define CLEAR            (1<<6)// CL bit (w07)
#define NO_CLEAR         0
#define LOAD             (1<<5)// LD bit (w06)
#define NO_LOAD          0
// LSC4434 command register address offsets
#define CHANMAX_OFS  8 		// RN bits (w09-w13)
#define CHANSTART_OFS 0         // FA bits (w01-w05)

// global variables
static int  N, qx, TLAM, slowmode, chanmax, chanstart, shots;

/******************* function(s) common to all channels **********************/
int lsc4434_readout ( Scheduler *parameters )
{
  int           d;
  unsigned int  dummy;

  // sequential data readout with autoincrement of address
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(2)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(2,0,N,&dummy)) < 3) {
    printf("\nlsc4434: F(2)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: No Q response "))
      return 0;
  }
  d = (int) dummy;
#ifdef DEBUG_DEV_LSC4434
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  parameters->store((parameters->act_channel() % shots == 0)?(d):(-1));  //save measured or no-meas value

#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: %s data stored (%d) ",(parameters->act_channel() % shots == 0)?("real"):("virtual"),d);
#endif

  return 1;
}
/*****************************************************************************/

//init function
//called each time before measurement starts
int lsc4434_init(List *device_data, Scheduler *parameters)
{	
  unsigned int dummy;

  if (camac_init() != 1)
    if (!parameters->error(E_CAMAC,"Couldn't initialize device "))
      return 0;

  printf("\ninitializing lsc4434");
  device_data->getvar(&N,"lsc4434","N");
  device_data->getvar(&slowmode,"lsc4434","slow_mode");
  device_data->getvar(&shots,"lsc4434","shots");
  device_data->getvar(&chanstart,"lsc4434","startchannel");
  device_data->getvar(&chanmax,"lsc4434","channelmax");
  device_data->getvar(&TLAM,"all","timeout");

  // clear command register flags and all 32 scalers
  dummy = NO_TEST | AUXBUS_DISABLE | (31<<CHANMAX_OFS) | READOUT_DISABLE | CLEAR | NO_LOAD | (0<<CHANSTART_OFS);
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(16)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(16,0,N,&dummy)) < 3) {
    printf("\nlsc4434: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: Couldnt load command register "))
      return 0;
  }
#ifdef DEBUG_DEV_LSC4434
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  return 1;
}

int lsc4434_pre ( List *device_data, Scheduler *parameters )
{
  unsigned int dummy;
  
  // prepare readout: set parameters for sequential readout and clear scalers
  dummy = NO_TEST | AUXBUS_DISABLE | ((chanmax-1)<<CHANMAX_OFS) | READOUT_DISABLE | CLEAR | NO_LOAD | (chanstart<<CHANSTART_OFS);
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(16)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(16,0,N,&dummy)) < 3) {
    printf("\nlsc4434: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: Couldnt load command register "))
      return 0;
  }
#ifdef DEBUG_DEV_LSC4434
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif
  return 1;
}

int lsc4434 ( List *device_data, Scheduler *parameters )
{
  unsigned int dummy;
           int i;

  // request data readout (will generate LAM if there is data to be read)
  dummy = NO_TEST | AUXBUS_DISABLE | ((chanmax-1)<<CHANMAX_OFS) | READOUT_ENABLE | NO_CLEAR | NO_LOAD | (chanstart<<CHANSTART_OFS);
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(16)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(16,0,N,&dummy)) < 3) {
    printf("\nlsc4434: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: Couldnt load command register "))
      return 0;
  }
#ifdef DEBUG_DEV_LSC4434
  printf("-> QX: %d\tD: %d ",qx,dummy);
#endif
  
  // wait for LAM
  i = 0;
  dummy = 0;
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(8)A(0)[%d] ",dummy);
#endif
  qx = camac(8,0,N,&dummy);  // test for LAM pending
  while (++i < TLAM && !(qx & 2))
  {
    if (slowmode) usleep(1);
    qx = camac(8,0,N,&dummy);
  }

#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434_%d: qx %d, TLAM %d/%d",channel,qx,i,TLAM);	
#endif
	
  if (i >= TLAM)  // no LAM occured - timeout failure
  {
    if (!parameters->error(E_TIMEOUT,"LeCroy LSC4434"))
      return 0;   //return FALSE -> measurement stops
  }
  
  // after LAM occured in time - load scalers into buffers (latches)
  dummy = NO_TEST | AUXBUS_DISABLE | ((chanmax-1)<<CHANMAX_OFS) | READOUT_ENABLE | NO_CLEAR | LOAD | (chanstart<<CHANSTART_OFS);
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(16)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(16,0,N,&dummy)) < 3) {
    printf("\nlsc4434: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: Couldnt load command register "))
      return 0;
  }
#ifdef DEBUG_DEV_LSC4434
  printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  usleep(1);
  
  // proceed sequential data readout
  for (i = 0; i < chanmax; i++) {
    if ( !(lsc4434_readout(parameters)))
      return 0;
  }
  
  dummy = 0;
  camac(10,0,N,&dummy);  // (test and) clear LAM

  usleep(1);

  // check if LAM went on again (=> data readout not finished!)
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(10)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(10,0,N,&dummy)) > 1) { // LAM still alive
    printf("\nlsc4434_readout(): F(10)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: Data readout not completed "))
      return 0;
  }
#ifdef DEBUG_DEV_LSC4434
  printf("-> QX: %d\tD: %d ",qx,dummy);
#endif
  
  return 1;
}

int lsc4434_close(List *device_data, Scheduler *parameters)
{
  unsigned int dummy;

  printf("\nclosing lsc4434 ");
  
  // clear command register flags and all 32 scalers
  dummy = NO_TEST | AUXBUS_DISABLE | (31<<CHANMAX_OFS) | READOUT_DISABLE | CLEAR | NO_LOAD | (0<<CHANSTART_OFS);
#ifdef DEBUG_DEV_LSC4434
  printf("\nlsc4434: F(16)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(16,0,N,&dummy)) < 3) {
    printf("\nlsc4434: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"lsc4434: Couldnt load command register "))
      return 0;
  }
#ifdef DEBUG_DEV_LSC4434
  printf("-> QX: %d\tD: %d ",qx,dummy);
#endif
    
  // reset CAMAC dataway
  if (camac_close() != 1)
    if (!parameters->error(E_CAMAC,"Couldn't close device"))
      return 0;

  return 1; //function has to return a value -> 1 otherwise measurement stops
}

