/*
 * MAX : tcd2277_gated.cc, Wed Jan  2 14:44:59 CET 2002 -sg
 *
 * Device driver for for the LeCroy time-to-digital converter TDC 2277
 * 	IN COMBINATION with the 2277/EXP2 hit expander
 *	gated version (software timed gates)
 *
 * This file is part of Max data acquisition software
 * (c) 2000 by M. Dammasch
 *
 * Max is free software; you can redistribute it and/or modify it
 * under the terms of the version 2 of GNU General Public License as
 * published by the Free Software Foundation.
 *
 * Max is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * (see the file COPYING) in this directory; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "list.hh"
#include "sched.hh"
#include "camacs.h"
#include "error.hh"

#undef DEBUG_DEV_TDC2277GATED

#define TDC2277GATED_CHMAX  65536

/*
  TDC2277 command register flags:
*/
// aquisition mode: bit 16 of status word
#define COMMON_START  (1<<15)
#define COMMON_STOP   0
//common start timeout: bits 13-15 of status word
#define COMMON_START_TIMEOUT_FP_CONTROLLED  0
#define COMMON_START_TIMEOUT_01024NS  ((1)<<12)
#define COMMON_START_TIMEOUT_02048NS  ((2)<<12)
#define COMMON_START_TIMEOUT_04096NS  ((3)<<12)
#define COMMON_START_TIMEOUT_08192NS  ((4)<<12)
#define COMMON_START_TIMEOUT_16384NS  ((5)<<12)
#define COMMON_START_TIMEOUT_32768NS  ((6)<<12)
#define COMMON_START_TIMEOUT_65536NS  ((7)<<12)
// hit leading edge: bit 12 of status word
#define HIT_EDGE_LEADING_ENABLED  (1<<11)
#define HIT_EDGE_LEADING_DISABLED  0
// hit trailing edge: bit 11 of status word
#define HIT_EDGE_TRAILING_ENABLED  (1<<10)
#define HIT_EDGE_TRAILING_DISABLED  0
// test input channels: bits 9-10 of status word
#define TEST_INPUT_NONE   0
#define TEST_INPUT_ODD   (1<<8)
#define TEST_INPUT_EVEN  (2<<8)
#define TEST_INPUT_ALL   (3<<8)
// trigger mode: bit 8 of status word
#define TRIGGER_MODE_LATCH  (1<<7)
#define TRIGGER_MODE_PULSE   0
// bits 5-7: unused!
// tester hit duration: bits 3-4 of status word
#define TEST_HIT_DURATION_0128NS   0
#define TEST_HIT_DURATION_0256NS  (1<<2)
#define TEST_HIT_DURATION_1024NS  (2<<2)
#define TEST_HIT_DURATION_2048NS  (3<<2)
// tester hit burst: bits 1-2 of status word
#define TEST_HIT_BURST_1  0
#define TEST_HIT_BURST_2  1
#define TEST_HIT_BURST_4  2
#define TEST_HIT_BURST_8  3
/***/

typedef struct {
  int start;
  int stop;
  int sum;
} TDC_Gate;

// global variables
static      int  N, qx, TLAM, chmax, gatemax, *hitarray;
static TDC_Gate *gate;

//init function
//called each time before measurement starts
int tdc2277gated_init(List *device_data, Scheduler *parameters)
{	
  unsigned int dummy;
  char         txt[32];
  int          i;

  if (camac_init() != 1)
    if (!parameters->error(E_CAMAC,"Couldn't initialize device "))
      return 0;

  printf("\ninitializing tdc2277gated");
  device_data->getvar(&TLAM,"all","timeout");
  device_data->getvar(&N,"tdc2277gated","N");
  device_data->getvar(&chmax,"tdc2277gated","channelmax");
  device_data->getvar(&gatemax,"tdc2277gated","gatemax");

  if (chmax > TDC2277GATED_CHMAX)
    chmax = TDC2277GATED_CHMAX;

  // allocate hitarray
  if ((hitarray = ((int *)calloc(chmax,sizeof(int)))) == NULL) {
    printf("\ntdc2277gated: Not enough memory for hit-array (%d ints)\n",chmax);
    if (!parameters->error(E_SYS_MEM,"tdc2277gated: Not enough memory for 'hitarray'"))
      return 0;
  }

  // allocate gatearray
  if ((gate = ((TDC_Gate *)calloc(gatemax,sizeof(TDC_Gate)))) == NULL) {
    printf("\ntdc2277gated: Not enough memory for gate-array (%d gates)\n",gatemax);
    if (!parameters->error(E_SYS_MEM,"tdc2277gated: Not enough memory for 'gatearray'"))
      return 0;
  }

  // set gates
  for (i = 0; i < gatemax; i++) {
    sprintf(txt,"gate%02d_start",i);
    device_data->getvar(&gate[i].start,"tdc2277gated",txt);
    sprintf(txt,"gate%02d_stop",i);
    device_data->getvar(&gate[i].stop,"tdc2277gated",txt);
  }

  // clear module
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(9)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(9,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(9)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt clear device "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  // set program status word (command register)
  dummy = COMMON_START |COMMON_START_TIMEOUT_FP_CONTROLLED | HIT_EDGE_LEADING_ENABLED | HIT_EDGE_TRAILING_DISABLED | TEST_INPUT_NONE | TRIGGER_MODE_PULSE | TEST_HIT_DURATION_0128NS | TEST_HIT_BURST_1;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(17)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(17,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(17)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt set program status word "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  // disable LAM
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(24)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(24,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(24)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt disable LAM "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  return 1;
}

int tdc2277gated_pre ( List *device_data, Scheduler *parameters )
{
  unsigned int dummy;

  // enable LAM
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(26)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(26,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(26)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt enable LAM "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  return 1;
}

int tdc2277gated ( List *device_data, Scheduler *parameters )
{
  unsigned int dummy;
           int i, bip, time, hitphase, channel, b;

  // wait for LAM
  i = 0;
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(8)A(0)[%d] ",dummy);
#endif
  qx = camac(8,0,N,&dummy);  // test for LAM pending
  while (++i < TLAM && !(qx & 2))
  {
    usleep(1);
    qx = camac(8,0,N,&dummy);
  }

#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: qx %d, TLAM %d/%d",qx,i,TLAM);	
#endif
	
  if (i >= TLAM) { // no LAM occured - timeout failure
    if (!parameters->error(E_TIMEOUT,"LeCroy TDC2277GATED"))
      return 0;   //return FALSE -> measurement stops
  }

  // clear LAM
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(10)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(10,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(10)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt clear LAM "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

    // wait 4 microsecs before readout starts
    usleep(4);

  // after LAM occured in time - read out hit data
    memset(hitarray,0,chmax*sizeof(int));

  i = 0;
  bip = 1;

  while (bip) {
    dummy = 0;
    if ( ! ((qx = camac(0,0,N,&dummy)) & 1)) {  // no x
      printf("\ntdc2277gated: F(0)A(0)[%d] -> QX: %d ",dummy,qx);
      if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt read data "))
	return 0;
    }
    while (qx & 2) { // data is valid
      time = dummy & 65535;
      hitphase = dummy >> 16;
      channel = dummy >> 17;
      if (chmax*time/TDC2277GATED_CHMAX < chmax-8) {  // discard internal timeout hits
	i++;
	hitarray[chmax*time/TDC2277GATED_CHMAX]++;
#ifdef DEBUG_DEV_TDC2277GATED
      printf("\ntdc2277gated: Read hit data #%003d: CH%02d: %d (%s edge) -> Bin#%d ",i,channel,time,(hitphase?"leading":"trailing"),chmax*time/TDC2277GATED_CHMAX);
#endif
      }
      dummy = 0;
      if ( ! ((qx = camac(0,0,N,&dummy)) & 1)) {  // no x
        printf("\ntdc2277gated: F(0)A(0)[%d] -> QX: %d ",dummy,qx);
        if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt read data "))
	  return 0;
      }
    }
    if ( ! (qx & 2)) {
      dummy = 0;
      if ( ! ((qx = camac(27,0,N,&dummy)) & 1)) {  // no x
        printf("\ntdc2277gated: F(27)A(0)[%d] -> QX: %d ",dummy,qx);
        if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt read BIP status "))
	  return 0;
      }
      bip = qx & 2;
    }
  }

  device_data->putvar(i,"tdc2277gated","hitcount");  // export actual hitrate (last integration)
  
  //parameters->store(hitarray,chmax);        // save the hit data array

  for (i = 0; i < gatemax; i++) {
    gate[i].sum = 0;
    for (b = gate[i].start; b <= gate[i].stop; b++) {
      gate[i].sum += hitarray[b];
    }
    parameters->store(gate[i].sum);
  }

  
  // clear module
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(9)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(9,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(9)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt clear device "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  // disable LAM
  dummy = 0;
#ifdef DEBUG_DEV_TDC2277GATED
  printf("\ntdc2277gated: F(24)A(0)[%d] ",dummy);
#endif
  if ((qx = camac(24,0,N,&dummy)) < 3) {
    printf("\ntdc2277gated: F(24)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_CAMAC,"tdc2277gated: Couldnt disable LAM "))
      return 0;
  }
#ifdef DEBUG_DEV_TDC2277GATED
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  return 1;
}

int tdc2277gated_close(List *device_data, Scheduler *parameters)
{
  printf("\nclosing tdc2277gated ");

  free(hitarray);
  free(gate);

  // reset CAMAC dataway
  if (camac_close() != 1)
    if (!parameters->error(E_CAMAC,"Couldn't close device"))
      return 0;

  return 1; //function has to return a value -> 1 otherwise measurement stops
}

