/*
 * MAX : camacs.cc, Sam Dez 29 19:24:57 CET 2001
 * 
 * CAMAC interface for use with the MAX data aquisition program
 *
 * Two CAMAC interface-controller systems are supported:
 * 
 * -  CAMAC Functions for the LeCroy 8901 GPIB Crate controller: _CAMAC_GPIB_
 *
 *    11.12.1988  M.Voelkel
 *    31.01.1994  D.Rohloff    geaendert fuer microsys ieee
 *    04.01.1996  J.Ludwig     Programmierstil geaendert
 *    10.07.1997  C.Otto       fr Linux umgeschrieben
 *    08.04.1998  M. Dammasch  auf mehrere IEEE-Devices verallgemeinert
 *
 * - CAMAC Functions for the Wiener CC16 CAMAC Crate controller: _CAMAC_CC16_
 *
 *   This driver is based on the c111a driver for the Wiener PC16Turbo card
 *   (c) Giovanni Organtini, organtini@roma1.infn.it
 *
 *   08.09.1998  M. Dammasch
 */

// This version is for Linux only

#include <stdio.h>
#include <stdlib.h>

#ifdef _CAMAC_GPIB_
#include <ib.h>
#include "gpiberr.h"
#endif

#ifdef _CAMAC_CC16_
#include <c111a.h>
#include <naf.h>
#endif

#include "camacs.h"

/*
 * global variables
 */
#ifdef _CAMAC_GPIB_
int  Eid_Cam;  // GPIB-Address of LeCroy-CAMAC-Controller
#endif
#ifdef _CAMAC_CC16_
int  cc16_fd;  // file descriptor of CC16-CAMAC-Controller
#endif
int  CamInit;  // CAMAC initialization flag

int  camac_init(void) 
{
#ifdef _CAMAC_GPIB_
  char d[5];
#endif
#ifdef _CAMAC_CC16_
  int q, x, data;
#endif

  if (CamInit) // exit if already initialized
    return 1;

  CamInit = 0; // set CAMAC not initialized

#ifdef _CAMAC_GPIB_
  printf("\n***** Connecting GPIB Device 'LeCroy' ***** ");

  // ibfind aus Lib sucht in /etc/gpib.conf nach "Geraet" und gibt Adresse zurck
  if((Eid_Cam = ibfind("LeCroy")) & ERR){
    gpiberr("ibfind error");
    return 0;
  }
  printf("\n'LeCroy' GPIB-Address is set to #%d ",Eid_Cam);    

  printf("\nsend IFC ");  // ibsic aus Lib.
  if( ibsic(Eid_Cam) & ERR){
    gpiberr("ibsic err");
    return 0;
  }
  printf("done\n");

  printf("send REM ");  // ibsre aus Lib.
  if( ibsre(Eid_Cam,1) & ERR){
    gpiberr("ibsre err");
    return 0;
  }
  printf("done\n");

  printf("clearing GPIB bus for 'LeCroy' ... ");  // ibclr aus Lib.
  if( ibclr(Eid_Cam) & ERR){
    gpiberr("ibclear error");
    return 0;
  }
  printf("done\n");
#endif
  
#ifdef _CAMAC_CC16_
  printf("\n\n***** Connecting CAMAC-Controller 'CC16' ***** ");
  
  if ((cc16_fd = c111aopen(CC16_ADDR)) <= 0) {  // failed
    switch (cc16_fd) {
    case -2 :
      fprintf(stderr,"\nCamac_Init(): Couldn't get file handle for CC16@%d ",CC16_ADDR); 
      break;
    case -1 :
      fprintf(stderr,"\nCamac_Init(): Invalid crate address (%d) for CC16 ",CC16_ADDR); 
      break;
    default :
      fprintf(stderr,"\nCamac_Init(): Illegal file descriptor (%d) for CC16@%d ",cc16_fd,CC16_ADDR); 
      break;
    }
    return 0;
  }
  printf("\n'CC16' file handle is set to #%d ",cc16_fd);
#endif
  
  /****** Init Crate Controller *******/
#ifdef _CAMAC_GPIB_

  printf("LeCroy: send CAMAC-Clear(C) and CAMAC-Initialize(Z) ... ");
  if( ibwrt (Eid_Cam,"\x23",1)& ERR ){  // Send both CLEAR and INITIALIZE
    gpiberr("write error");
    return 0;
  }
  if( ibrd (Eid_Cam,d,4)& ERR){
    gpiberr("read error");
    return 0;	
  }
  printf("done\n");
  
  printf("LeCroy: set data mode: ");
  if( ibwrt (Eid_Cam,"\x64",1)& ERR ){  // set data transfer mode (24bit)
    gpiberr("write error");
    return 0;        
  }
  if( ibrd (Eid_Cam,d,4)  & ERR){
    gpiberr("read error");
    return 0;
  }
  printf("24-bit read normal transfer ");
#endif

#ifdef _CAMAC_CC16_
  q = x = data = 0;
  printf("\nCC16_%d.25: F(36)A(1)[%d]: Clear I, C cycle ",cc16_fd,data);
  c111a(cc16_fd, 25, 1, 36, &q, &x, &data);
  if (x == 1)  // okay
    printf("done\n");
  else {
    printf("failed (QX: %d <%d%d>\t D: %d (0x%x)) ",2*q+x,q,x,data,data); 
    fprintf(stderr,"\nCamac_Init(): CC16_%d.25: Couldn't execute F(36)A(1) ",cc16_fd);
    return 0;
  }

  q = x = data = 0;
  printf("\nCC16_%d.25: F(38)A(1)[%d]: Clear I, Z, C cycle ",cc16_fd,data);
  c111a(cc16_fd, 25, 1, 38, &q, &x, &data);
  if (x == 1)  // okay
    printf("done\n");
  else {
    printf("failed (QX: %d <%d%d>\t D: %d (0x%x)) ",2*q+x,q,x,data,data); 
    fprintf(stderr,"\nCamac_Init(): CC16_%d.25: Couldn't execute F(38)A(1) ",cc16_fd);
    return 0;
  }
#endif

  CamInit = 1;
  printf("\n***** Completed Successfully *****\n");

  return 1;
 }


int camac_close ( void )
{
#ifdef _CAMAC_GPIB_
  char d[5];
#endif
#ifdef _CAMAC_CC16_
  int q, x, data;
#endif
  
  if (CamInit)
  { 
#ifdef _CAMAC_GPIB_
    printf("\nLeCroy: send CAMAC-Clear(C) ... ");
    if( ibwrt (Eid_Cam,"\x22",1)& ERR ){  // Send CLEAR 
      gpiberr("write error");
      return 0;
    }
    if( ibrd (Eid_Cam,d,4)& ERR){
      gpiberr("read error");
      return 0;	
    }
    printf("done\n");  

    printf("clearing GPIB bus for 'LeCroy' ... ");  // ibclr aus Lib.
    if( ibclr(Eid_Cam) & ERR){
      gpiberr("ibclear error");
      return 0;
    }
    printf("done\n");
#endif
    
#ifdef _CAMAC_CC16_
    q = x = data = 0;
    printf("\nCC16_%d.25: F(38)A(1)[0]: Clear I, Z, C cycle ",cc16_fd);
    c111a(cc16_fd, 25, 1, 38, &q, &x, &data);
    if (x == 1)  // okay
      printf("done\n");
    else {
      printf("failed (QX: %d <%d%d>\t D: %d (0x%x)) ",2*q+x,q,x,data,data); 
      fprintf(stderr,"\nCamac_Close(): CC16_%d.25: Couldn't execute F(38)A(1) ",cc16_fd);
      return 0;
    }
    
    c111aclose(cc16_fd);
#endif
  }
  
  CamInit = 0;

  return 1;
}


int camac(int f,int a,int n,unsigned int *d) /* Deklaration mit Int noetig da auch */
{                                     /* ein Zeiger uebergeben werden kann  */

#ifdef _CAMAC_GPIB_
  char camac_befehl[6];      /* Befehlsstring fuer CAMAC Aufruf */
  char camac_daten[4];       /* Datenstring fuer CAMAC Antwort */
  int anzahl_bytes;                   /* Anzahl der Bytes fuer Uebertragung */

  if (f > CAMAC_BEFEHL)
    {                               /* CAMAC Befehl ungueltig */
      printf("Wrong CAMAC setting due to\n");
      exit(0);
    }
  else                              /* CAMAC Befehl gueltig */
    {
      camac_befehl[CAMAC_F] = (char) (f & 0xff);     /* Unteres Byte von f in Befehlsstring */
      camac_befehl[CAMAC_N] = (char) (n & 0xff);     /* Unteres Byte von n in Befehlsstring */
      camac_befehl[CAMAC_A] = (char) (a & 0xff);     /* Unteres Byte von a in Befehlsstring */
      
      switch (f & CAMAC_BEFEHLS_MASKE)
	{
	case CAMAC_READ:
	case CAMAC_SET_STATUS_A:
	case CAMAC_SET_STATUS_B:
	  anzahl_bytes = 3;
	  break;
	case CAMAC_WRITE:
	  camac_befehl[CAMAC_D1] = (char) (*d & 0xff);
	  camac_befehl[CAMAC_D2] = (char) ((*d & 0xff00 ) >> 8);
	  camac_befehl[CAMAC_D3] = (char) ((*d & 0xff0000 ) >> 16);
	  anzahl_bytes = 6;
	  break;
	default:
	  anzahl_bytes = 0;
	  printf("Wrong CAMAC setting due to\n");
	  exit(0);
	  break;
	} /* of switch f */
      
      /* CAMAC Befehl schreiben ibwrt aus Lib. */
      if( ibwrt(Eid_Cam, camac_befehl, anzahl_bytes) & ERR ){
	printf("\n Warning, error occured!");
	gpiberr("write error");
	exit(1);
      }
      
      /* CAMAC Antwort lesen (immer 4 Bytes) ibrd aus Lib.  */
      if( ibrd (Eid_Cam, camac_daten, 4) & ERR){
	printf("\n Warning, error occured!");
	gpiberr("read error");
	exit(1);
      }
#ifdef DEBUG
      printf("\nCAMAC@%02d.%02d: F(%02d)A(%02d)[%d] -> QX: %d   D: %d ",Eid_Cam,n,f,a,*d,(int)camac_daten[3],(int) (camac_daten[0] | ( camac_daten[1] << 8 ) | ( camac_daten[2] << 16 )));
#endif
             
      if ((f & CAMAC_BEFEHLS_MASKE) == CAMAC_READ)
	{                                           /* Daten zusaetzlich zu QX wurden gelesen */
   	 *d=camac_daten[0] |
	    ( camac_daten[1] << 8 ) |
	    ( camac_daten[2] << 16 ) ;
	}
      return ( (int) camac_daten[3]);   /* Rckgabewert: QX (=3 bei Erfolg)*/
    } /* of f <= CAMAC BEFEHL */
#endif
  
#ifdef _CAMAC_CC16_
  int q=0, x=0;
  
#ifdef DEBUG
  printf("\nCC16_%d.%02d: F(%d)A(%d)[%d] ",cc16_fd,n,f,a,*d);
#endif      
  
  c111a(cc16_fd, n, a, f, &q, &x,(int *) d);
  
#ifdef DEBUG
  printf("-> QX: %d <%d%d>\t D: %d (0x%x) ",2*q+x,q,x,*d,*d);
#endif               
  
  return(2*q+x);
#endif
}
